"""Mock for offline testing of `FlowController`s."""
from __future__ import annotations

import asyncio
import contextlib
from random import choice, random
from unittest.mock import AsyncMock, MagicMock

from .basis import GASES as BASISGases
from .driver import CONTROL_POINTS, GASES, MAX_RAMP_TIME_UNITS, MaxRampTimeUnit
from .util import Client as RealClient


class Client(RealClient):
    """Mock the alicat communication client."""

    def __init__(self, address: str) -> None:
        super().__init__(timeout=0.01)
        _ = address
        self.writer = MagicMock(spec=asyncio.StreamWriter)
        self.writer.write.side_effect=self._handle_write
        self.reader = AsyncMock(spec=asyncio.StreamReader)
        self.reader.read.return_value = self.eol
        self.reader.readuntil.side_effect = self._handle_read

        self.port = address
        self.open = True
        self._next_reply = ''

        self.open = True
        self.control_point: str = choice(list(CONTROL_POINTS))
        self.state: dict[str, str | float] = {
            'setpoint': 10,
            'gas': 'N2',
            'mass_flow': 10 * (0.95 + 0.1 * random()),
            'pressure': random() * 50.0,
            'temperature': random() * 50.0,
            'total_flow': 0.0,
            'volumetric_flow': 0.0,
        }
        self.ramp_config = { 'up': False, 'down': False, 'zero': False, 'power': False }
        self.button_lock: bool = False
        self.keys = ['pressure', 'temperature', 'volumetric_flow', 'mass_flow',
                     'setpoint', 'gas']
        self.firmware = '6v21.0-R22 Nov 30 2016,16:04:20'
        self.max_ramp_time_unit : MaxRampTimeUnit

    async def _handle_connection(self) -> None:
        pass

    def _create_dataframe(self) -> str:
        """Generate a typical 'dataframe' with current operating conditions."""
        return (
            f"{self.unit} "
            f"{self.state['pressure']:+07.2f} "
            f"{self.state['temperature']:+07.2f} "
            f"{self.state['volumetric_flow']:+07.2f} "
            f"{self.state['mass_flow']:+07.2f} "
            f"{self.state['setpoint']:07.2f} "
            f"{self.state['gas']:<7} "
            f"{'LCK' if self.button_lock else ''}"
        )
    def _create_ramp_response(self) -> str:
        """Generate a response to setting or getting the ramp config."""
        config = self.ramp_config
        return (f"{self.unit}"
                f" {1 if config['up'] else 0}"
                f" {1 if config['down'] else 0}"
                f" {1 if config['zero'] else 0}"
                f" {1 if config['power'] else 0}")

    def _create_max_ramp_response(self) -> str:
        """Generate a response to setting or getting the max ramp rate."""
        return (f"{self.unit}"
                f" {self.max_ramp:.7f}"
                f" 7"  # SLPM  # fixme make dynamic
                f" {MAX_RAMP_TIME_UNITS[self.max_ramp_time_unit]}"
                f" SLPM/{self.max_ramp_time_unit}")

    def _handle_write(self, data: bytes) -> None:
        """Act on writes sent to the mock client, updating internal state and setting self._next_reply if necessary."""
        msg = data.decode().strip()
        self.unit = msg[0]

        msg = msg[1:]  # strip unit
        if msg == '':  # get dataframe
            self._next_reply = self._create_dataframe()
        elif msg == '$$L':  # lock
            self.button_lock = True
            self._next_reply = self._create_dataframe()
        elif msg == '$$U':  # unlock
            self.button_lock = False
            self._next_reply = self._create_dataframe()
        elif 'W122=' in msg:  # set control point
            cp = int(msg[5:])
            self.control_point = next(p for p, i in CONTROL_POINTS.items() if cp == i)
            self._next_reply = f"{self.unit}   122 = {cp}"
        elif msg == 'R122':  # read control point
            self._next_reply = f"{self.unit}   122 = {CONTROL_POINTS[self.control_point]}"
        elif msg == 'LSRC':  # get ramp config
            self._next_reply = self._create_ramp_response()
        elif 'LSRC' in msg:  # set ramp config
            values = msg[5:].split(' ')
            self.ramp_config = {
                'up': values[0] == '1',
                'down': values[1] == '1',
                'zero': values[2] == '1',
                'power': values[3] == '1',
            }
            self._next_reply = self._create_ramp_response()
        elif msg == 'SR':  # get max ramp rate
            self._next_reply = self._create_max_ramp_response()
        elif 'SR' in msg:  # set max ramp rate
            values = msg.split()
            self.max_ramp = float(values[1])
            unit_time_int = int(values[2])
            self.max_ramp_time_unit = next(key for key, val in MAX_RAMP_TIME_UNITS.items() if val == unit_time_int)
            self._next_reply = self._create_max_ramp_response()
        elif msg[0] == 'S':  # set setpoint
            self.state['setpoint'] = float(msg[1:])
            self._next_reply = self._create_dataframe()
        elif msg[0:6] == '$$W46=':  # set gas via reg46
            gas = msg[6:]
            self._next_reply = f"{self.unit}   046 = {gas}"
            with contextlib.suppress(ValueError):
                gas = GASES[int(gas)]
            self.state['gas'] = gas
        elif msg == '$$R46':  # read gas via reg46
            gas_index = GASES.index(self.state['gas'])  # type: ignore
            reg46_value = gas_index & 0x1FF  # encode gas number in low 9 bits
            self._next_reply = f"{self.unit}   046 = {reg46_value}"
        elif msg == 'VE':  # get firmware
            self._next_reply = self.firmware
        elif msg == '$$PC':
            self.state['pressure'] = 0
            self._next_reply = self._create_dataframe()
        elif msg == '$$V':
            self.state['volumetric_flow'] = 0
            self._next_reply = self._create_dataframe()
        else:
            raise NotImplementedError(msg)

    async def _handle_read(self, separator: bytes) -> bytes:
        """Reply to read requests from the mock client."""
        reply = self._next_reply.encode() + separator
        self._next_reply = ''
        return reply

class BASISClient(RealClient):
    """Mock the BASIS communication client."""

    def __init__(self, address: str, baudrate: int) -> None:
        super().__init__(timeout=0.01)
        self.writer = MagicMock(spec=asyncio.StreamWriter)
        self.writer.write.side_effect=self._handle_write
        self.reader = AsyncMock(spec=asyncio.StreamReader)
        self.reader.read.return_value = self.eol
        self.reader.readuntil.side_effect = self._handle_read

        self.port = address
        self.baud = baudrate
        self.open = True
        self._next_reply = ''

        self.open = True
        self.state: dict[str, str | float] = {
            'setpoint': 10,
            'gas': 'N2',
            'mass_flow': 10 * (0.95 + 0.1 * random()),
            'temperature': random() * 50.0,
            'valve_drive': 0.0,
            'control_point': 'mass_flow',
            'totalizer': 0.0,
        }
        self.keys = ['temperature', 'mass_flow', 'totalizer',
                     'valve_drive', 'gas', 'setpoint', 'control_point']
        self.pid = {'P': '500', 'I': '5000'}
        self.batchvol = '0'
        self.firmware = "V 3.1.0"

    async def _handle_connection(self) -> None:
        pass

    def _create_dataframe(self) -> str:
        """Generate a typical 'dataframe' with current operating conditions."""
        return (
            f"{self.unit} "
            f"{self.state['temperature']:+07.2f} "
            f"{self.state['mass_flow']:+07.2f} "
            f"{self.state['totalizer']:+07.2f} "
            f"{self.state['setpoint']:07.2f} "
            f"{self.state['valve_drive']:07.2f} "
            f"{self.state['gas']:<7} "
            f"{'HLD' if self.state['control_point'] == 'HLD' else ''}"
        )

    def _handle_write(self, data: bytes) -> None:
        """Act on writes sent to the mock client, updating internal state and setting self._next_reply if necessary."""
        msg = data.decode().strip()
        self.unit = msg[0]

        msg = msg[1:]  # strip unit
        if msg == '':  # get dataframe
            self._next_reply = self._create_dataframe()
        elif 'VE' in msg: # get firmware
            self._next_reply = f"{self.unit} {self.firmware}"
        elif msg[0] == 'S':  # set setpoint
            self.state['setpoint'] = float(msg[1:])
            self._next_reply = self._create_dataframe()
        elif "GS" in msg:  # set gas via reg46
            gas = msg[3:]
            self._next_reply = f"{self.unit} {gas}"
            with contextlib.suppress(ValueError):
                gas = BASISGases[int(gas)]
            self.state['gas'] = gas
        elif msg[0] == 'V': # tare flow
            self.state['mass_flow'] = 0
            self._next_reply = self._create_dataframe()
        elif msg == 'T': # reset totalizer
            self.state['totalizer'] = 0
            self._next_reply = self._create_dataframe()
        elif 'HPUR' in msg: # valve hold
            self.state['control_point'] = 'HLD'
            valve_drive = float(msg[5:])
            self.state['valve_drive'] = valve_drive
            self._next_reply = self._create_dataframe()
        elif msg == 'C': # cancel hold
            self.state['control_point'] = 'mass_flow'
            self._next_reply = self._create_dataframe()
        elif 'LCG' in msg: # pid terms
            if len(msg.strip().split(' ')) == 1: # get pid terms
                self._next_reply = f"{self.unit} {self.pid['P']} {self.pid['I']}"
            else:
                terms = msg[3:].strip().split(' ')
                self.pid = {'P': terms[0], 'I': terms[1]}
                self._next_reply = f"{self.unit} {msg[3:]}"
        elif 'TB' in msg: ## batch volume
            if len(msg.strip().split(' ')) == 1: # get batch vol
                self._next_reply = f"{self.batchvol}"
            else:
                self.batchvol = msg[2:]
                self._next_reply = f"{self.unit} {self.batchvol}"
        elif 'DV 64' in msg: ## remaining volume
            self._next_reply = f"{self.unit} {msg[3:]}"
        else:
            raise NotImplementedError(msg)

    async def _handle_read(self, separator: bytes) -> bytes:
        """Reply to read requests from the mock client."""
        reply = self._next_reply.encode() + separator
        self._next_reply = ''
        return reply



"""Python driver for Alicat mass flow controllers, using serial communication.

Distributed under the GNU General Public License v2
Copyright (C) 2024 Alex Ruddick
Copyright (C) 2023 NuMat Technologies
"""
from __future__ import annotations

import asyncio
from typing import Any, ClassVar, Literal

from .util import Client, SerialClient, TcpClient, _is_float

GASES = ['Air', 'Ar', 'CH4', 'CO', 'CO2', 'C2H6', 'H2', 'He',
         'N2', 'N2O', 'Ne', 'O2', 'C3H8', 'n-C4H10', 'C2H2',
         'C2H4', 'i-C2H10', 'Kr', 'Xe', 'SF6', 'C-25', 'C-10',
         'C-8', 'C-2', 'C-75', 'A-75', 'A-25', 'A1025', 'Star29',
         'P-5']
class FlowMeter:
    """Python driver for Alicat Flow Meters.

    [Reference](http://www.alicat.com/
    products/mass-flow-meters-and-controllers/mass-flow-meters/).

    This communicates with the flow meter over a USB or RS-232/RS-485
    connection using pyserial, or an Ethernet <-> serial converter.
    """

    # mapping of port names to a tuple of Client objects and their refcounts
    open_ports: ClassVar[dict[str, tuple[Client, int]]] = {}

    def __init__(self, address: str = '/dev/ttyUSB0', unit: str = 'A', **kwargs: Any) -> None:
        """Connect this driver with the appropriate USB / serial port.

        Args:
            address: The serial port or TCP address:port. Default '/dev/ttyUSB0'.
            unit: The Alicat-specified unit ID, A-Z. Default 'A'.
        """
        if address.startswith('/dev') or address.startswith('COM'):  # serial
            if address in FlowMeter.open_ports:
                # Reuse existing connection
                self.hw, refcount = FlowMeter.open_ports[address]
                FlowMeter.open_ports[address] = (self.hw, refcount + 1)
            else:
                # Open a new connection and store it
                self.hw: Client = SerialClient(address=address, **kwargs)  # type: ignore[no-redef]
                FlowMeter.open_ports[address] = (self.hw, 1)
        else:
            self.hw = TcpClient(address=address, **kwargs)

        self.unit = unit
        self.keys = ['pressure', 'temperature', 'volumetric_flow', 'mass_flow',
                     'setpoint', 'gas']
        self.open = True
        self.firmware: str | None = None

    async def __aenter__(self) -> FlowMeter:
        """Provide async enter to context manager."""
        return self

    async def __aexit__(self, *args: Any) -> None:
        """Provide async exit to context manager."""
        await self.close()
        return

    @classmethod
    async def is_connected(cls, port: str, unit: str = 'A') -> bool:
        """Return True if the specified port is connected to this device.

        This class can be used to automatically identify ports with connected
        Alicats. Iterate through all connected interfaces, and use this to
        test. Ports that come back True should be valid addresses.

        Note that this distinguishes between `FlowController` and `FlowMeter`.
        """
        is_device = False
        try:
            device = cls(port, unit)
            try:
                c = await device.get()
                if cls.__name__ == 'FlowMeter':
                    assert c
                    assert 'setpoint' not in device.keys
                elif cls.__name__ == 'FlowController':
                    assert c
                    assert 'setpoint' in device.keys
                else:
                    raise NotImplementedError('Must be meter or controller.')
                is_device = True
            finally:
                await device.close()
        except Exception:
            pass
        return is_device

    def _test_controller_open(self) -> None:
        """Raise an IOError if the FlowMeter has been closed.

        Does nothing if the meter is open and good for read/write
        otherwise raises an IOError. This only checks if the meter
        has been closed by the FlowMeter.close method.
        """
        if not self.open:
            raise OSError(f"The FlowMeter with unit ID {self.unit} and "
                          f"port {self.hw.address} is not open")

    async def _write_and_read(self, command: str) -> str | None:
        """Wrap the communicator request, to call _test_controller_open() before any request."""
        self._test_controller_open()
        return await self.hw.write_and_read(command)

    async def get(self) -> dict[str, Any]:
        """Get the current state of the flow controller.

        From the Alicat mass flow controller documentation, this data is:
         * Pressure (normally in psia)
         * Temperature (normally in C)
         * Volumetric flow (in units specified at time of order)
         * Mass flow (in units specified at time of order)
         * Total flow (only on models with the optional totalizer function)
         * Currently selected gas

        Args:
            retries: Number of times to re-attempt reading. Default 2.
        Returns:
            The state of the flow controller, as a dictionary.

        """
        command = f'{self.unit}'
        line = await self._write_and_read(command)
        if not line:
            raise OSError("Could not read values")
        spl = line.split()
        unit, values = spl[0], spl[1:]

        # Over range errors for mass, volume, pressure, and temperature
        # Explicitly silenced because I find it redundant.
        while values[-1].upper() in ['MOV', 'VOV', 'POV', 'TOV']:
            del values[-1]
        if unit != self.unit:
            raise ValueError("Flow controller unit ID mismatch.")
        if values[-1].upper() == 'LCK':
            self.button_lock = True
            del values[-1]
        else:
            self.button_lock = False
        if len(values) == 5 and len(self.keys) == 6:
            del self.keys[-2]
        elif len(values) == 7 and len(self.keys) == 6:
            self.keys.insert(5, 'total flow')
        elif len(values) == 2 and len(self.keys) == 6:
            self.keys.insert(1, 'setpoint')
        elif len(values) == 4 and len(self.keys) == 6:  # LCR (liquid)
            del self.keys[-1]  # gas
            del self.keys[2]  # volumetric flow
        return {k: (float(v) if _is_float(v) else v)
                for k, v in zip(self.keys, values, strict=True)}
    async def set_gas(self, gas: str | int) -> None:
        """Set the gas type.

        Args:
            gas: The gas type, as a string or integer. Supported strings are:
                'Air', 'Ar', 'CH4', 'CO', 'CO2', 'C2H6', 'H2', 'He', 'N2',
                'N2O', 'Ne', 'O2', 'C3H8', 'n-C4H10', 'C2H2', 'C2H4',
                'i-C2H10', 'Kr', 'Xe', 'SF6', 'C-25', 'C-10', 'C-8', 'C-2',
                'C-75', 'A-75', 'A-25', 'A1025', 'Star29', 'P-5'

                Gas mixes may only be called by their mix number.
        """
        if isinstance(gas, str):
            if gas not in GASES:
                raise ValueError(f"{gas} not supported!")
            gas_number = GASES.index(gas)
        else:
            gas_number = gas
        command = f'{self.unit}$$W46={gas_number}'
        # fixme does this overwrite the upper bits??
        _ = await self._write_and_read(command)
        reg46 = await self._write_and_read(f'{self.unit}$$R46')
        if not reg46:
            raise OSError("Cannot set gas.")
        reg46_gasbit = int(reg46.split()[-1]) & 0b0000000111111111

        if gas_number != reg46_gasbit:
            raise OSError("Cannot set gas.")

    async def create_mix(self, mix_no: int, name: str, gases: dict[str, float]) -> None:
        """Create a gas mix.

        Gas mixes are made using COMPOSER software.
        COMPOSER mixes are only allowed for firmware 5v or greater.

        Args:
        mix_no: The mix number. Gas mixes are stored in slots 236-255.
        name: A name for the gas that will appear on the front panel.
        Names greater than six letters will be cut off.
        gases: A dictionary of the gas by name along with the associated
        percentage in the mix.
        """
        firmware = await self.get_firmware()
        if any(v in firmware for v in ['2v', '3v', '4v', 'GP']):
            raise OSError("This unit does not support COMPOSER gas mixes.")

        if mix_no < 236 or mix_no > 255:
            raise ValueError("Mix number must be between 236-255!")

        total_percent = sum(gases.values())
        if total_percent != 100:
            raise ValueError("Percentages of gas mix must add to 100%!")

        if any(gas not in GASES for gas in gases):
            raise ValueError("Gas not supported!")

        gas_list = [f'{percent} {GASES.index(gas)}' for gas, percent in gases.items()]
        command = ' '.join([
            self.unit,
            'GM',
            name,
            str(mix_no),
            ' '.join(gas_list),
        ])
        line = await self._write_and_read(command)

        # If a gas mix is not successfully created, ? is returned.
        if line == '?':
            raise OSError("Unable to create mix.")

    async def delete_mix(self, mix_no: int) -> None:
        """Delete a gas mix."""
        command = f'{self.unit}GD{mix_no}'
        line = await self._write_and_read(command)
        if line == '?':
            raise OSError("Unable to delete mix.")

    async def lock(self) -> None:
        """Lock the buttons."""
        command = f'{self.unit}$$L'
        _ = await self._write_and_read(command)

    async def unlock(self) -> None:
        """Unlock the buttons."""
        command = f'{self.unit}$$U'
        _ = await self._write_and_read(command)

    async def is_locked(self) -> bool:
        """Return whether the buttons are locked."""
        _ = await self.get()
        return self.button_lock

    async def tare_pressure(self) -> None:
        """Tare the pressure."""
        command = f'{self.unit}$$PC'
        line = await self._write_and_read(command)
        if line == '?':
            raise OSError("Unable to tare pressure.")

    async def tare_volumetric(self) -> None:
        """Tare volumetric flow."""
        command = f'{self.unit}$$V'
        line = await self._write_and_read(command)
        if line == '?':
            raise OSError("Unable to tare flow.")

    async def reset_totalizer(self) -> None:
        """Reset the totalizer."""
        command = f'{self.unit}T'
        _ = await self._write_and_read(command)

    async def get_firmware(self) -> str:
        """Get the device firmware version."""
        if self.firmware is None:
            command = f'{self.unit}VE'
            self.firmware = await self._write_and_read(command)
        if not self.firmware:
            raise OSError("Unable to get firmware.")
        return self.firmware

    async def flush(self) -> None:
        """Read all available information. Use to clear queue."""
        self._test_controller_open()
        await self.hw.clear()

    async def close(self) -> None:
        """Close the flow meter. Call this on program termination.

        Also close the serial port if no other FlowMeter object has
        a reference to the port.
        """
        if not self.open:
            return
        port = self.hw.address
        if port in FlowMeter.open_ports:
            connection, refcount = FlowMeter.open_ports[port]
            if refcount > 1:
                FlowMeter.open_ports[port] = (connection, refcount - 1)
            else:
                await connection.close()  # Close the port if no other instance uses it
                del FlowMeter.open_ports[port]
        self.open = False

CONTROL_POINTS = {
    'mass flow': 37, 'vol flow': 36,
    'abs pressure': 34, 'gauge pressure': 38, 'diff pressure': 39,
}  # fixme: add remaining control points
MaxRampTimeUnit = Literal['ms', 's', 'm', 'h', 'd']
MAX_RAMP_TIME_UNITS: dict[MaxRampTimeUnit, int] = {
    'ms': 3,
    's': 4,
    'm': 5,
    'h': 6,
    'd': 7,
}

class FlowController(FlowMeter):
    """Python driver for Alicat Flow Controllers.

    [Reference](http://www.alicat.com/products/mass-flow-meters-and-
    controllers/mass-flow-controllers/).

    This communicates with the flow controller over a USB or RS-232/RS-485
    connection using pyserial.

    To set up your Alicat flow controller, power on the device and make sure
    that the "Input" option is set to "Serial".
    """

    def __init__(self, address: str='/dev/ttyUSB0', unit: str='A', **kwargs: Any) -> None:
        """Connect this driver with the appropriate USB / serial port.

        Args:
            address: The serial port or TCP address:port. Default '/dev/ttyUSB0'.
            unit: The Alicat-specified unit ID, A-Z. Default 'A'.
        """
        FlowMeter.__init__(self, address, unit, **kwargs)
        self.control_point: str | None = None
        async def _init_control_point() -> None:
            self.control_point = await self._get_control_point()
        self._init_task = asyncio.create_task(_init_control_point())

    async def __aenter__(self) -> FlowController:
        """Provide async enter to context manager."""
        return self

    async def _write_and_read(self, command: str) -> str | None:
        """Wrap the communicator request.

        (1) Ensure _init_task is called once before the first request
        (2) Call _test_controller_open() before any request
        """
        if 'R122' not in command:
            await self._init_task
        self._test_controller_open()
        return await self.hw.write_and_read(command)

    async def get(self) -> dict[str, Any]:
        """Get the current state of the flow controller.

        From the Alicat mass flow controller documentation, this data is:
         * Pressure (normally in psia)
         * Temperature (normally in C)
         * Volumetric flow (in units specified at time of order)
         * Mass flow (in units specified at time of order)
         * Flow setpoint (in units of control point)
         * Flow control point (e.g. 'mass flow' or 'abs pressure')
         * Total flow (only on models with the optional totalizer function)
         * Currently selected gas

        Returns:
            The state of the flow controller, as a dictionary.
        """
        state = await super().get()
        state['control_point'] = self.control_point
        return state

    async def set_flow_rate(self, flowrate: float) -> None:
        """Set the target flow rate.

        Args:
            flow: The target flow rate, in units specified at time of purchase
        """
        if self.control_point in ['abs pressure', 'gauge pressure', 'diff pressure']:
            await self._set_setpoint(0)
            await self._set_control_point('mass flow')
        await self._set_setpoint(flowrate)

    async def set_pressure(self, pressure: float) -> None:
        """Set the target pressure.

        Args:
            pressure: The target pressure, in units specified at time of
                purchase. Likely in psia.
        """
        if self.control_point in ['mass flow', 'vol flow']:
            await self._set_setpoint(0)
            await self._set_control_point('abs pressure')
        await self._set_setpoint(pressure)

    async def get_totalizer_batch(self, batch: int = 1) -> str:
        """Get the totalizer batch volume (firmware 10v00).

        Args:
            batch: Which of the two totalizer batches to query.
                Default is 1; some devices have 2

        Returns:
            line: Current value of totalizer batch
        """
        command = f'{self.unit}$$TB {batch}'
        line = await self._write_and_read(command)
        if line == '?':
            raise OSError("Unable to read totalizer batch volume.")
        values = line.split(" ")  # type: ignore[union-attr]
        return f'{values[2]} {values[4]}' # returns 'batch vol' 'units'

    async def set_totalizer_batch(self, batch_volume: float, batch: int = 1, units: str = 'default') -> None:
        """Set the totalizer batch volume (firmware 10v00).

        Args:
            batch: Which of the two totalizer batches to set.
                Default is 1; some devices have 2
            batch_volume: Target batch volume, in same units as units
                on device
            units: Units of the volume being provided. Default
                is 0, so device returns default engineering units.
        """
        engineering_units_table = {"default":0, "SμL":2, "SmL":3, "SL":4, \
                    "Scm3":6, "Sm3":7, "Sin3":8, "Sft3":9, "kSft3":10, "NμL":32, \
                    "NmL":33, "NL":34, "Ncm3":36, "Nm3":37}

        if units in engineering_units_table:
            units_no = engineering_units_table[units]
        else:
            raise ValueError("Units not in unit list. Please consult Appendix B-3 of the Alicat Serial Primer.")

        command = f'{self.unit}$$TB {batch} {batch_volume} {units_no}'
        line = await self._write_and_read(command)
        if line == '?':
            raise OSError("Unable to set totalizer batch volume. Check if volume is out of range for device.")

    async def hold(self) -> None:
        """Override command to issue a valve hold (firmware 5v07).

        For a single valve controller, hold the valve at the present value.
        For a dual valve flow controller, hold the valve at the present value.
        For a dual valve pressure controller, close both valves.
        """
        command = f'{self.unit}$$H'
        _ = await self._write_and_read(command)

    async def cancel_hold(self) -> None:
        """Cancel valve hold."""
        command = f'{self.unit}$$C'
        _ = await self._write_and_read(command)

    async def get_pid(self) -> dict[str, Any]:
        """Read the current PID values on the controller.

        Values include the loop type, P value, D value, and I value.
        Values returned as a dictionary.
        """
        self.pid_keys = ['loop_type', 'P', 'D', 'I']

        command = f'{self.unit}$$r85'
        read_loop_type = await self._write_and_read(command)
        if not read_loop_type:
            raise OSError("Could not get PID values.")
        spl = read_loop_type.split()

        loopnum = int(spl[3])
        loop_type = ['PD/PDF', 'PD/PDF', 'PD2I'][loopnum]
        pid_values = [loop_type]
        for register in range(21, 24):
            value = await self._write_and_read(f'{self.unit}$$r{register}')
            if not value:
                raise OSError(f"Could not read register {register}")
            value_spl = value.split()
            pid_values.append(value_spl[3])

        return {k: (v if k == self.pid_keys[-1] else str(v))
                for k, v in zip(self.pid_keys, pid_values, strict=True)}

    async def set_pid(self, p: int | None=None,
                            i: int | None=None,
                            d: int | None=None,
                            loop_type: str | None=None) -> None:
        """Set specified PID parameters.

        Args:
            p: Proportional gain
            i: Integral gain. Only used in PD2I loop type.
            d: Derivative gain
            loop_type: Algorithm option, either 'PD/PDF' or 'PD2I'

        This communication works by writing Alicat registers directly.
        """
        if loop_type is not None:
            options = ['PD/PDF', 'PD2I']
            if loop_type not in options:
                raise ValueError(f'Loop type must be {options[0]} or {options[1]}.')
            loop_num=options.index(loop_type) + 1
            command = f'{self.unit}$$w85={loop_num}'
            _ = await self._write_and_read(command)
        if p is not None:
            command = f'{self.unit}$$w21={p}'
            _ = await self._write_and_read(command)
        if i is not None:
            command = f'{self.unit}$$w23={i}'
            _ = await self._write_and_read(command)
        if d is not None:
            command = f'{self.unit}$$w22={d}'
            _ = await self._write_and_read(command)

    async def _set_setpoint(self, setpoint: float) -> None:
        """Set the target setpoint.

        Called by `set_flow_rate` and `set_pressure`, which both use the same
        command once the appropriate register is set.
        """
        command = f'{self.unit}S{setpoint:.2f}'
        line = await self._write_and_read(command)
        if not line:
            raise OSError("Could not set setpoint.")
        try:
            current = float(line.split()[5])
        except IndexError:
            current = None
        if current is not None and abs(current - setpoint) > 0.01:
            # possibly the setpoint is being ramped
            command = f'{self.unit}LS'
            line = await self._write_and_read(command)
            if not line:
                raise OSError("Could not set setpoint.")
            try:
                commanded = float(line.split()[2])
            except IndexError:
                raise OSError("Could not set setpoint.") from None
            if commanded is not None and abs(commanded - setpoint) > 0.01:
                raise OSError("Could not set setpoint.")

    async def _get_control_point(self) -> str:
        """Get the control point, and save to internal variable."""
        command = f'{self.unit}R122'
        line = await self._write_and_read(command)
        if not line:
            raise OSError("Could not read control point.")
        value = int(line.split('=')[-1])
        try:
            cp = next(p for p, r in CONTROL_POINTS.items() if value == r)
            self.control_point = cp
            return cp
        except StopIteration:
            raise ValueError(f"Unexpected register value: {value:d}") from None

    async def _set_control_point(self, point: str) -> None:
        """Set the variable used as the control point.

        Args:
            point: 'mass flow', 'vol flow', 'abs pressure', 'gauge pressure', or 'diff pressure'
        """
        if point not in CONTROL_POINTS:
            raise ValueError(f"Control point must be one of {list(CONTROL_POINTS.keys())}.")
        reg = CONTROL_POINTS[point]
        command = f'{self.unit}W122={reg:d}'
        line = await self._write_and_read(command)
        if not line:
            raise OSError("Could not set control point.")
        value = int(line.split('=')[-1])
        if value != reg:
            raise OSError("Could not set control point.")
        self.control_point = point

    async def set_ramp_config(self, config: dict[str, bool]) -> None:
        """Configure the setpoint ramp settings (firmware 10v05).

        `up`: whether the controller ramps when increasing the setpoint,
        `down`: whether the controller ramps when decreasing the setpoint,
                (this includes setpoints below 0 on bidirectional devices),
        `zero`: whether the controller ramps when establishing a zero setpoint,
        `power`: whether the controller ramps when using a power-up setpoint
        """
        command = (f"{self.unit}LSRC"
                  f" {1 if config['up'] else 0}"
                  f" {1 if config['down'] else 0}"
                  f" {1 if config['zero'] else 0}"
                  f" {1 if config['power'] else 0}")
        line = await self._write_and_read(command)
        if not line or self.unit not in line:
            raise OSError("Could not set ramp config.")


    async def get_ramp_config(self) -> dict[str, bool]:
        """Get the setpoint ramp settings (firmware 10v05).

        `up`: whether the controller ramps when increasing the setpoint,
        `down`: whether the controller ramps when decreasing the setpoint,
                (this includes setpoints below 0 on bidirectional devices),
        `zero`: whether the controller ramps when establishing a zero setpoint,
        `power`: whether the controller ramps when using a power-up setpoint
        """
        command = f"{self.unit}LSRC"
        line = await self._write_and_read(command)
        if not line or self.unit not in line:
            raise OSError("Could not read ramp config.")
        values = line[2:].split(' ')
        if len(values) != 4:
            raise OSError("Could not read ramp config.")
        return {
            'up': values[0] == '1',
            'down': values[1] == '1',
            'zero': values[2] == '1',
            'power': values[3] == '1',
        }

    async def set_maxramp(self, max_ramp: float,
                          unit_time: MaxRampTimeUnit) -> None:
        """Set the maximum ramp rate (firmware 7v11).

        Args:
            max_ramp: The maximum ramp rate
            unit_time: The units of the ramp rate
                - 3: (m)illisecond
                - 4: (s)econd
                - 5: (m)inute
                - 6: (h)hour
                - 7: (d)ay
        """
        command = f"{self.unit}SR {max_ramp:.2f} {MAX_RAMP_TIME_UNITS[unit_time]}"
        line = await self._write_and_read(command)
        if not line or self.unit not in line:
            raise OSError("Could not set max ramp.")

    async def get_maxramp(self) -> dict[str, float | str]:
        """Get the maximum ramp rate (firmware 7v11).

        Returns:
            max_ramp: The maximum ramp rate
            units: The units string returned from the controller
        """
        command = f"{self.unit}SR"
        line = await self._write_and_read(command)
        if not line or self.unit not in line:
            raise OSError("Could not read max ramp.")
        values = line.split(' ')
        if len(values) != 5:
            raise OSError("Could not read max ramp.")
        return {
            'max_ramp': float(values[1]),
            'units': str(values[4]),
        }



"""Python driver for Alicat BASIS mass flow devices, using serial communication.

Distributed under the GNU General Public License v2
"""
from __future__ import annotations

from typing import Any

from .driver import FlowMeter
from .util import Client, SerialClient, _is_float

GASES = ['Air', 'Ar', 'CO2', 'N2', 'O2', 'N2O', 'H2', 'He', 'CH4']

class BASISMeter(FlowMeter):
    """Python driver for BASIS Flow Meters.

    [Reference](https://www.alicat.com/wp-content/documents/manuals/DOC-MANUAL-BASIS2.pdf).

    This communicates with the flow meter over a USB or RS-232/RS-485
    connection using pyserial.
    """

    def __init__(self, address: str = '/dev/ttyUSB0', unit: str = 'A', baudrate: int = 38400, **kwargs: Any) -> None:
        """Connect this driver with the appropriate USB / serial port.

        Args:
            address: The serial port or TCP address:port. Default '/dev/ttyUSB0'.
            unit: The Alicat-specified unit ID, A-Z. Default 'A'.
            baudrate: The baud rate of the device. Default 38400.
        """
        self.hw: Client = SerialClient(address=address, baudrate=baudrate, **kwargs)

        self.unit = unit
        self.keys = ['temperature', 'mass_flow', 'totalizer', 'setpoint',
                     'valve_drive', 'gas']
        self.open = True
        self.firmware: str | None = None

    async def __aenter__(self, *args: Any) -> BASISMeter:
        """Provide async enter to context manager."""
        return self

    @classmethod
    async def is_connected(cls, port: str, unit: str = 'A') -> bool:
        """Return True if the specified port is connected to this device.

        This class can be used to automatically identify ports with connected
        Alicats. Iterate through all connected interfaces, and use this to
        test. Ports that come back True should be valid addresses.
        """
        is_device = False
        try:
            device = cls(port, unit)
            try:
                c = await device.get()
                assert c
                is_device = True
            finally:
                await device.close()
        except Exception:
            pass
        return is_device

    async def get(self) -> dict[str, Any]:
        """Get the current state of the flow controller.

        From the Alicat mass flow controller documentation, this data is:
         * Pressure (normally in psia)
         * Temperature (normally in C)
         * Mass flow (in units specified at time of order)
         * Total flow
         * Currently selected gas

        Args:
            retries: Number of times to re-attempt reading. Default 2.
        Returns:
            The state of the flow controller, as a dictionary.

        """
        command = f'{self.unit}'
        line = await self._write_and_read(command)
        if not line:
            raise OSError("Could not read values")
        spl = line.split()
        unit, values = spl[0], spl[1:]

        # Over range errors for mass, volume, pressure, and temperature
        # Explicitly silenced because I find it redundant.
        while values[-1].upper() in ['MOV', 'VOV', 'POV', 'TOV']:
            del values[-1]
        if unit != self.unit:
            raise ValueError("Flow controller unit ID mismatch.")
        if len(values) == 5 and len(self.keys) == 6:
            self.keys.remove('setpoint')
        return {k: (float(v) if _is_float(v) else v)
                for k, v in zip(self.keys, values, strict=True)}

    async def set_gas(self, gas: str | int) -> None:
        """Set the gas type.

        Args:
            gas: The gas type, as a string or integer. Supported strings are:
                'Air', 'Ar', 'CO2', 'N2', 'O2', 'N2O', 'H2', 'He', 'CH4'
        """
        if isinstance(gas, str):
            if gas not in GASES:
                raise ValueError(f"{gas} not supported!")
            gas_number = GASES.index(gas)
        else:
            gas_number = gas
        command = f'{self.unit}GS {gas_number}'
        res = await self._write_and_read(command)

        if not res:
            raise OSError("Cannot set gas.")

    async def tare(self, duration: int = 10) -> None:
        """Tare flow.

        Args:
            duration: The time in ms to collect values for calculating the tare
            offset. The value is a positive integer from 1-32767.
        """
        command = f'{self.unit}V {duration}'
        line = await self._write_and_read(command)

        if line == '?':
            raise OSError("Unable to tare flow.")

    async def reset_totalizer(self) -> None:
        """Reset the totalizer."""
        command = f'{self.unit}T'
        await self._write_and_read(command)

    async def get_firmware(self) -> str:
        """Get the device firmware version."""
        if self.firmware is None:
            command = f'{self.unit}VE'
            self.firmware = await self._write_and_read(command)
        if not self.firmware:
            raise OSError("Unable to get firmware.")
        return self.firmware

    async def flush(self) -> None:
        """Read all available information. Use to clear queue."""
        self._test_controller_open()
        await self.hw.clear()

    async def close(self) -> None:
        """Close the flow meter. Call this on program termination.

        Also closes the serial port if no other FlowMeter object has
        a reference to the port.
        """
        if not self.open:
            return
        await self.hw.close()
        self.open = False


class BASISController(BASISMeter):
    """Python driver for Alicat Flow Controllers.

    [Reference](http://www.alicat.com/products/mass-flow-meters-and-
    controllers/mass-flow-controllers/).

    This communicates with the flow controller over a USB or RS-232/RS-485
    connection using pyserial.

    To set up your Alicat flow controller, power on the device and make sure
    that the "Input" option is set to "Serial".
    """

    def __init__(self, address: str='/dev/ttyUSB0', unit: str='A', baudrate: int = 38400, **kwargs: Any) -> None:
        """Connect this driver with the appropriate USB / serial port.

        Args:
            address: The serial port. Default '/dev/ttyUSB0'.
            unit: The Alicat-specified unit ID, A-Z. Default 'A'.
            baudrate: The baud rate of the device. Default 38400.
        """
        BASISMeter.__init__(self, address, unit, baudrate, **kwargs)

    async def __aenter__(self, *args: Any) -> BASISController:
        """Provide async enter to context manager."""
        return self

    async def get(self) -> dict[str, Any]:
        """Get the current state of the flow controller.

        From the Alicat mass flow controller documentation, this data is:
         * Temperature (normally in C)
         * Mass flow (in units specified at time of order)
         * Flow setpoint (in units of control point)
         * Flow control point (either 'mass_flow' or 'HLD')
         * Total flow
         * Currently selected gas

        Returns:
            The state of the flow controller, as a dictionary.
        """
        cp = "mass flow"

        command = f'{self.unit}'
        line = await self._write_and_read(command)
        if not line:
            raise OSError("Could not read values")
        spl = line.split()
        unit, values = spl[0], spl[1:]

        # Over range errors for mass, volume, pressure, and temperature
        # Explicitly silenced because I find it redundant.
        while values[-1].upper() in ['MOV', 'VOV', 'POV', 'TOV']:
            del values[-1]
        if unit != self.unit:
            raise ValueError("Flow controller unit ID mismatch.")
        if len(values) == 5 and len(self.keys) == 6:
            del self.keys[-3]
        if values[-1] == "HLD":
            cp = "HLD"
            del values[-1]
        state = {k: (float(v) if _is_float(v) else v)
                for k, v in zip(self.keys, values, strict=False)}
        state['control_point'] = cp
        return state

    async def get_totalizer_batch(self) -> list[float]:
        """Get the totalizer batch volume.

        Returns:
            volume: Batch volume
            remaining: Remaining batch volume
        """
        remaining = await self._write_and_read(f'{self.unit}DV 64')
        current = await self._write_and_read(f'{self.unit}TB')
        if current == '?' or remaining == '?':
            raise OSError("Unable to read totalizer batch volume.")
        return [current, remaining.split(" ")[-1]] # type: ignore

    async def set_totalizer_batch(self, batch_volume: float) -> None:
        """Set the totalizer batch volume.

        Args:
            batch_volume: Target batch volume, in same units as units
                on device
        """
        command = f'{self.unit}TB {batch_volume}'
        line = await self._write_and_read(command)

        if line == '?':
            raise OSError("Unable to set totalizer batch volume. Check if volume is out of range for device.")

    async def hold(self, percentage: float) -> None:
        """Override command to issue a valve hold at a certain percentage of full drive.

        Args:
            percentage : Percentage of full valve drive
        """
        command = f'{self.unit}HPUR {percentage}'
        await self._write_and_read(command)

    async def cancel_hold(self) -> None:
        """Cancel valve hold."""
        command = f'{self.unit}C'
        await self._write_and_read(command)

    async def get_pid(self) -> dict[str, str]:
        """Read the current PID values on the controller.

        Values include the P value and I value.
        Values returned as a dictionary.
        """
        self.pid_keys = ['P', 'I']

        command = f'{self.unit}LCG'
        line = await self._write_and_read(command)
        if not line:
            raise OSError("Could not get PID values.")
        spl = line.split()
        return dict(zip(self.pid_keys, spl[1:], strict=False))

    async def set_pid(self, p: int, i: int) -> None:
        """Set specified PID parameters.

        Args:
            p: Proportional gain
            i: Integral gain. Only used in PD2I loop type.
        """
        command = f'{self.unit}LCG {p} {i}'
        await self._write_and_read(command)

    async def set_flow_rate(self, flowrate: float) -> None:
        """Set the target setpoint."""
        command = f'{self.unit}S {flowrate}'
        line = await self._write_and_read(command)
        if not line:
            raise OSError("Could not set setpoint.")
        try:
            current = float(line.split()[4])
        except IndexError:
            raise OSError("Could not set setpoint.") from None
        if current is not None and abs(current - flowrate) > 0.1:
            raise OSError("Could not set setpoint.")



"""Test the driver responds with correct data."""
from random import uniform
from unittest import mock

import pytest

from alicat import command_line
from alicat.driver import FlowController
from alicat.mock import Client

ADDRESS = 'COM12'

@pytest.fixture(scope='module', autouse=True)
async def patch_serial_client(request):
    """Replace the serial client with our mock."""
    if request.config.getoption("--hardware"):
        async with FlowController(ADDRESS) as device:
            res = await device._write_and_read('A')
            if not res:
                pytest.exit("Device not found. Ensure device is connected " \
                "with unit id A and BAUD rate 19200 on the correct COM port.")
        yield
    else:
        with mock.patch('alicat.driver.SerialClient', Client):
            yield

@pytest.mark.parametrize('unit', ['A', 'B'])
def test_driver_cli(capsys, unit):
    """Confirm the commandline interface works with different unit IDs."""
    command_line([ADDRESS, '--unit', unit])
    captured = capsys.readouterr()
    assert ("mass_flow" in captured.out)


@pytest.mark.parametrize('cls', [FlowController])  # Fixme: fix FlowMeter
async def test_is_connected(cls):
    """Confirm that connection status works."""
    async with cls(ADDRESS) as device:
        assert await device.is_connected(ADDRESS)
        assert not await device.is_connected('bad_address')


async def test_flow_setpoint_roundtrip():
    """Confirm that setting/getting flowrates works."""
    async with FlowController(ADDRESS) as device:
        flow_sp = round(uniform(0.01, 0.1), 2)
        await device.set_flow_rate(flowrate=flow_sp)
        # assert flow_sp == await device.get_flow_rate()
        result = await device.get()
        assert flow_sp == result['setpoint']


async def test_lock_unlock():
    """Confirm that locking/unlocking the buttons works."""
    async with FlowController(ADDRESS) as device:
        await device.lock()
        assert await device.is_locked()
        await device.unlock()
        assert not await device.is_locked()


@pytest.mark.parametrize('gas', ['Air', 'H2'])
async def test_set_standard_gas_name(gas):
    """Confirm that setting standard gases by name works."""
    async with FlowController(ADDRESS) as device:
        await device.set_gas(gas)
        result = await device.get()
        assert gas == result['gas']
        with pytest.raises(ValueError, match='not supported'):
            await device.set_gas('methylacetylene-propadiene propane')


@pytest.mark.parametrize('gas', [('Air', 0), ('H2', 6)])
async def test_set_standard_gas_number(gas):
    """Confirm that setting standard gases by number works."""
    async with FlowController(ADDRESS) as device:
        await device.set_gas(gas[1])
        result = await device.get()
        assert gas[0] == result['gas']


@pytest.mark.skip
async def test_create_gas_mix():
    """Confirm creating custom gas mixes works."""
    pass


async def test_tare_pressure():
    """Confirm taring the absolute pressure works."""
    async with FlowController(ADDRESS) as device:
        await device.tare_pressure()
        result = await device.get()
        assert result['pressure'] == 0.0


async def test_tare_flow():
    """Confirm taring the flow works."""
    async with FlowController(ADDRESS) as device:
        await device.tare_volumetric()
        result = await device.get()
        assert result['volumetric_flow'] == 0.0


async def test_get_firmware():
    """Confirm the firmware version can be read."""
    async with FlowController(ADDRESS) as device:
        result = await device.get_firmware()
        assert 'v' in result or 'GP' in result


@pytest.mark.parametrize('config', [
    {'up': True, 'down': False, 'zero': True, 'power': False},
    {'up': True, 'down': True, 'zero': False, 'power': True},
    {'up': False, 'down': False, 'zero': False, 'power': False},
    ])
async def test_ramp_config(config):
    """Confirm changing the ramping configuration works."""
    async with FlowController(ADDRESS) as device:
        await device.set_ramp_config(config)
        result = await device.get_ramp_config()
        assert config == result

@pytest.mark.parametrize('unit_time', ['ms', 's', 'm', 'h', 'd'])
async def test_maxramp(unit_time):
    """Confirm that setting/getting the maximum ramp rate works."""
    async with FlowController(ADDRESS) as device:
        max_ramp = round(uniform(0.01, 0.1), 2)
        await device.set_maxramp(max_ramp, unit_time)
        result = await device.get_maxramp()
        assert max_ramp == result['max_ramp']
        assert result['units'] == f'SLPM/{unit_time}'  # fixme make units dynamic

@pytest.mark.parametrize('control_point',
    ['mass flow', 'vol flow', 'abs pressure', 'gauge pressure', 'diff pressure'])
async def test_control_point(control_point):
    """Confirm changing the control point works."""
    async with FlowController(ADDRESS) as device:
        await device._set_control_point(control_point)
        result = await device._get_control_point()
        assert control_point == result



"""Test the BASIS devices/driver respond with correct data."""
from random import uniform
from unittest import mock

import pytest

from alicat.basis import BASISController, BASISMeter
from alicat.mock import BASISClient

ADDRESS = "COM16" # tests requite unit: A, baud: 38400

@pytest.fixture(scope='module', autouse=True)
async def patch_serial_client(request):
    """Replace the serial client with our mock."""
    if request.config.getoption("--hardware"):
        async with BASISMeter(ADDRESS) as device:
            res = await device._write_and_read('A')
            if not res:
                pytest.exit("Device not found. Ensure device is connected " \
                "with unit id A and BAUD rate 38400 on the correct COM port.")
        yield
        async with BASISMeter(ADDRESS) as device:
            await device._write_and_read('AS 0')
    else:
        with mock.patch('alicat.basis.SerialClient', BASISClient):
            yield

@pytest.mark.skip
@pytest.mark.parametrize('cls', [BASISMeter, BASISController])  # Fixme
async def test_is_connected(cls):
    """Confirm that connection status works."""
    async with cls(ADDRESS) as device:
        assert await device.is_connected(ADDRESS)
        assert not await device.is_connected('bad_address')

async def test_tare_flow():
    """Confirm taring the flow works."""
    async with BASISMeter(ADDRESS) as device:
        await device.tare()
        result = await device.get()
        assert result['mass_flow'] == 0.0

async def test_reset_totalizer():
    """Confirm resetting the totalizer works."""
    async with BASISController(ADDRESS) as device:
        await device.reset_totalizer()
        result = await device.get()
        assert result['totalizer'] == 0.0

@pytest.mark.parametrize('gas', ['Air', 'H2'])
async def test_set_standard_gas_name(gas):
    """Confirm that setting standard gases by name works."""
    async with BASISController(ADDRESS) as device:
        await device.set_gas(gas)
        result = await device.get()
        assert gas == result['gas']
        with pytest.raises(ValueError, match='not supported'):
            await device.set_gas('methylacetylene-propadiene propane')

async def test_get_set_pid_terms():
    """Confirm PID terms are updated properly."""
    async with BASISController(ADDRESS) as device:
        p = round(uniform(100, 500))
        i = round(uniform(1000, 5000))
        await device.set_pid(p, i)
        result = await device.get_pid()
        assert {"P": f'{p}', "I": f'{i}'} == result

async def test_totalizer_batch_volume():
    """Confirm setting the totalizer batch volume works."""
    async with BASISController(ADDRESS) as device:
        batch_vol = round(uniform(1, 100))
        await device.set_totalizer_batch(batch_vol)
        result = await device.get_totalizer_batch()
        assert batch_vol == pytest.approx(float(result[0]))

async def test_valve_hold():
    """Confirm holding at a given valve drive percentage works."""
    async with BASISController(ADDRESS) as device:
        valve_drive = round(uniform(1, 100))
        await device.hold(valve_drive)
        result = await device.get()
        assert "HLD" in result['control_point']

        await device.cancel_hold()
        result = await device.get()
        assert "HLD" not in result['control_point']

@pytest.mark.parametrize('gas', [('Air', 0), ('Ar', 1)])
async def test_set_standard_gas_number(gas):
    """Confirm that setting standard gases by number works."""
    async with BASISController(ADDRESS) as device:
        await device.set_gas(gas[1])
        result = await device.get()
        assert gas[0] == result['gas']

async def test_flow_setpoint_roundtrip():
    """Confirm that setting/getting flowrates works."""
    async with BASISController(ADDRESS) as device:
        flow_sp = round(uniform(1, 100), 2)
        await device.set_flow_rate(flowrate=flow_sp)
        result = await device.get()
        assert flow_sp == pytest.approx(result['setpoint'], 0.1)
